Package: spreadr
Type: Package
Title: Simulating Spreading Activation in a Network
Version: 0.3.0
Authors@R: c(
    person("Cynthia","Siew", role = c("aut", "cre"),
      email = "cynsiewsq@gmail.com", ),
    person(c("Dirk","U."), "Wulff", role = "ctb",
      email = "dirk.wulff@gmail.com"),
    person(c("Ning","Yuan"), "Lee", role = "ctb",
      email = "ningyuan.sg@gmail.com"))
Description: The notion of spreading activation is a prevalent metaphor in the cognitive sciences. This package provides the tools for cognitive scientists and psychologists to conduct computer simulations that implement spreading activation in a network representation. The algorithmic method implemented in 'spreadr' subroutines follows the approach described in Vitevitch, Ercal, and Adagarla (2011, Frontiers), who viewed activation as a fixed cognitive resource that could spread among nodes that were connected to each other via edges or connections (i.e., a network). See Vitevitch, M. S., Ercal, G., & Adagarla, B. (2011). Simulating retrieval from a highly clustered network: Implications for spoken word recognition. Frontiers in Psychology, 2, 369. <doi:10.3389/fpsyg.2011.00369> and Siew, C. S. Q. (2019). spreadr: A R package to simulate spreading activation in a network. Behavior Research Methods, 51, 910-929. <doi: 10.3758/s13428-018-1186-5>.
License: GPL-3
Encoding: UTF-8
Depends: Rcpp (>= 0.12.5), R (>= 2.10)
Imports: Matrix, assertthat, igraph
Suggests: dplyr, fs, ggplot2, gganimate, ggraph, gifski, knitr,
        rmarkdown, testthat (>= 3.0.0)
LazyData: true
VignetteBuilder: knitr
LinkingTo: Rcpp
NeedsCompilation: yes
Packaged: 2026-05-27 07:01:32 UTC; cynthia
Author: Cynthia Siew [aut, cre],
  Dirk U. Wulff [ctb],
  Ning Yuan Lee [ctb]
Maintainer: Cynthia Siew <cynsiewsq@gmail.com>
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
URL: https://csqsiew.github.io/spreadr/
Repository: CRAN
Date/Publication: 2026-05-30 08:30:08 UTC
Built: R 4.4.3; x86_64-apple-darwin20; 2026-05-30 11:26:23 UTC; unix
Archs: spreadr.so.dSYM

# Generated by roxygen2: do not edit by hand

export(spreadr)
importFrom(Rcpp,sourceCpp)
importFrom(assertthat,assert_that)
importFrom(assertthat,has_name)
importFrom(assertthat,is.count)
importFrom(assertthat,is.flag)
importFrom(igraph,"V<-")
importFrom(igraph,E)
importFrom(igraph,V)
importFrom(igraph,as_adjacency_matrix)
importFrom(igraph,graph_from_adjacency_matrix)
importFrom(igraph,is_igraph)
useDynLib(spreadr)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# spreadr

<!-- badges: start -->

# spreadr <img src="man/figures/logo.png" align="right" height="139" alt="" />

<!-- badges: end -->

The notion of spreading activation is a prevalent metaphor in the
cognitive sciences. This package provides the tools for cognitive
scientists and psychologists to conduct computer simulations that
implement spreading activation in a network representation. The
algorithmic method implemented in *spreadr* subroutines follows the
approach described in Vitevitch, Ercal, and Adagarla (2011, Frontiers),
who viewed activation as a fixed cognitive resource that could spread
among nodes that were connected to each other via edges or connections
(i.e., a network). See Vitevitch, M. S., Ercal, G., & Adagarla, B.
(2011).

## Installation

You can install the stable version via:

``` r
install.packages("spreadr")
```

or the latest version (on the GitHub master branch):

``` r
# install.packages("pak")
pak::pak("csqsiew/spreadr")
```

If you encounter an error which looks like:

`Error: (converted from warning) package 'Rcpp' was built under R version x.y.z`

try setting `Sys.setenv(R_REMOTES_NO_ERRORS_FROM_WARNINGS="true")` first
[^1].

## Development Notes

### Install developer dependencies

There are some additional packages that need to be installed for the
*spreadr* developer, e.g. those used in the unit tests. To install these
developer dependencies, use:

``` r
devtools::install_dev_deps()
```

### Testing with the semantic priming simulation

By default, running the test suite will replicate most of the
simulations in the article describing *spreadr* [^2] except for the
last, *“Simulation Study 3: Semantic priming”*. That simulation requires
800 iterations each taking a few minutes, making it impractical for
testing.

However, there is an option to include the replication of the first
iteration of that simulation, with `retention = 0.8`. Simply set the
environment variable `TEST_SEMANTIC_PRIMING` to any non-empty string.
For example, within an interactive R session:

``` r
Sys.setenv(TEST_SEMANTIC_PRIMING="true")
devtools::test()
```

Or, through `R CMD`:

    $ cd ..
    $ R CMD build spreadr
    $ TEST_SEMANTIC_PRIMING=true R CMD check spreadr_x.y.z.tar.gz

replacing `x.y.z` with the version string.

### Testing with animated GIFs

There are some snapshot tests checking for the generation the animated
GIFs. By default, these tests are not run. This is because (i) we expect
GIF generation to differ trivially with development environments (so
that the tests fail, but not in meaningful ways), and (ii) the spreadr
results underlying the GIF is always tested before the GIF generation.

However, the GIF generation snapshot tests are still available for use.
Simply set the environment variable `TEST_ANIMATED_GIF` to any non-empty
string, just as with the semantic priming simulation tests. These tests
can be useful for you to visualise the spreadr results.

### Invalid ELF header

If you encounter an “invalid ELF header” error, as in:

> devtools::load_all() Loading spreadr Error in dyn.load(dllfile) :
> unable to load shared object
> ‘/home/ning/github/spreadr/src/spreadr.so’:
> /home/ning/github/spreadr/src/spreadr.so: invalid ELF header

Try calling `devtools::clean_dll` first [^3].

[^1]: <https://github.com/r-lib/remotes/issues/403#issuecomment-748181946>

[^2]: <https://doi.org/10.3758/s13428-018-1186-5>

[^3]: <https://github.com/r-lib/devtools/issues/2027#issuecomment-483691800>

#' Small example of a phonological network as an igraph object
#'
#' @format igraph object representing an unweighted undirected graph with 34
#'   vertices and 96 edges. There are no self-loops.
#' @source Ying, Chan & Vitevitch, Michael. (2009). The Influence of the
#'   Phonological Neighborhood Clustering Coefficient on Spoken Word
#'   Recognition. Journal of experimental psychology. Human perception and
#'   performance. 35. 1934-49. 10.1037/a0016902.
"pnet"

#' Small example of a phonological network as an adjacency matrix
#'
#' @format Adjacency matrix representing an unweighted undirected graph with
#'   34 vertices and 96 edges. There are no self-loops.
#' @source Ying, Chan & Vitevitch, Michael. (2009). The Influence of the
#'   Phonological Neighborhood Clustering Coefficient on Spoken Word
#'   Recognition. Journal of experimental psychology. Human perception and
#'   performance. 35. 1934-49. 10.1037/a0016902.
"pnetm"

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

create_mat_t <- function(mat, a_tm1, d, retention) {
    .Call('_spreadr_create_mat_t', PACKAGE = 'spreadr', mat, a_tm1, d, retention)
}

#' Simulate spreading activation in a network
#'
#' @param network Adjacency matrix or
#'   \code{\link[igraph:igraph-package]{igraph}} object representing the network
#'   in which to simulate spreading activation.
#' @param start_run Non-empty \code{\link[base]{data.frame}} with mandatory
#'   columns \emph{node}, \emph{activation}; and optional columns \emph{time}.
#'   If the \emph{time} column is present, \emph{activation} is added to
#'   \emph{node} at each \emph{time}. Otherwise, the \emph{activation}s are
#'   added to their corresponding nodes at \code{t = 0}.
#' @param decay Number from 0 to 1 (inclusive) representing the proportion of
#'   activation that is lost at each time step.
#' @param retention Number from 0 to 1 (inclusive) or a numeric vector of such
#'   numbers of length equals number of nodes in the network. This represents
#'   the proportion of activation that remains in the node (not spread) at each
#'   time step. Then, \code{1 - retention} of the activation at each node is
#'   spread to neighbouring nodes. If a numeric vector, retentions are assigned
#'   to nodes according to the order given by \code{V(network)} if
#'   \code{network} is an \code{\link[igraph:igraph-package]{igraph}} object or
#'   \code{nrow(network)} if \code{network} is an adjacency matrix.
#' @param suppress Number representing the maximum amount of activation in a
#'   node for it to be set to 0, at each time step.
#' @param time Positive non-zero integer, or \code{NULL}. If not \code{NULL},
#'   the number of time steps to simulate before stopping. Otherwise,
#'   stop with the \code{threshold_to_stop} parameter.
#' @param threshold_to_stop Number or \code{NULL}. If not \code{NULL}, stop the
#'   simulation only when all nodes have activation value less than
#'   \code{threshold_to_stop}. Otherwise, stop with the \code{time} parameter.
#' @param include_t0 Boolean flag indicating if activation at \code{t = 0}
#'   should be prepended to the output \code{data.frame}. This is \code{FALSE}
#'   by default for back-compatibility.
#' @param create_names Boolean flag indicating if nodes should be automatically
#'   named (\code{1:n}, where \code{n} is the number of nodes) in case they are
#'   missing.
#' @param never_stop Boolean flag indicating if the simulation should be stopped
#'   if there have been too many iterations (so that there might be an infinite
#'   loop).
#'
#' @return A \code{\link[base]{data.frame}} with \emph{node}, \emph{activation}
#'   and \emph{time} columns representing the spread of activation in the network
#'   over time.
#'
#' @details At least one of parameters \code{time} or \code{threshold_to_stop}
#'   must be non-\code{NULL}. If both are non-\code{NULL}, the simulation stops
#'   at the earliest time possible.
#'
#'   The simulation iterates like so: for every \code{i} in \code{[0, time]},
#'   \itemize{
#'     \item{Spread activation from node to node}
#'     \item{Decay the activation at each node by the proportion specified by
#'       \code{decay}}
#'     \item{Set the activation at nodes with activation less than
#'       \code{suppress} to \code{0}}
#'     \item{Add the activations in \code{start_run} with \code{time = i} to
#'       their corresponding nodes}
#'     \item{Save the activations at each node for output}
#'     \item{Check the terminating conditions \code{time} and
#'       \code{threshold_to_stop}. If any are satisfied, terminate the
#'       simulation.}
#'   }
#'
#' @examples
#' # make an adjacency matrix and randomly fill some cells with 1s
#' mat <- matrix(sample(c(0,1), 100, replace=TRUE), 10, 10)
#' diag(mat) <- 0 # remove self-loops
#' initial_df <- data.frame(node=1, activation=20, stringsAsFactors=FALSE)
#' results <- spreadr(mat, initial_df)
#'
#' head(results, 10)
#' tail(results, 10)
#'
#' @importFrom assertthat assert_that is.count is.flag has_name
#' @importFrom igraph as_adjacency_matrix graph_from_adjacency_matrix is_igraph V V<- E
#' @importFrom Rcpp sourceCpp
#' @useDynLib spreadr
#'
#' @export
spreadr <- function(
    network, start_run, retention=0.5, time=10, threshold_to_stop=NULL,
    decay=0, suppress=0, include_t0=FALSE, create_names=TRUE, never_stop=FALSE
  ){

  # is network an igraph or square matrix-like object?
  assert_that(is_igraph(network) ||
    length(dim(network)) == 2 && dim(network)[1] == dim(network)[2])
  # is start_run in the correct format?
  assert_that(is.data.frame(start_run))
  assert_that(has_name(start_run, "node"), has_name(start_run, "activation"))
  assert_that(  # time is non-negative integer, if it is provided
    !has_name(start_run, "time") ||
      all(sapply(start_run$time, function(x) x == 0 || is.count(x))))
  # is decay is a number between 0 and 1 inclusive?
  assert_that(decay >= 0 && decay <= 1)
  # is retention an appropriate number or numeric vector?
  assert_that(is.numeric(retention))
  n_nodes <- if (is_igraph(network)) length(V(network)) else nrow(network)
  assert_that(length(retention) == 1 || length(retention) == n_nodes)
  assert_that(all(retention >= 0 & retention <= 1))
  # are terminating conditions ok? (time and threshold_to_stop)
  assert_that(
    !(is.null(time) && is.null(threshold_to_stop)),
    msg="time and threshold_to_stop cannot both be NULL")
  assert_that(is.null(time) || is.count(time))
  assert_that(is.null(threshold_to_stop) || threshold_to_stop > 0)
  # are node names ok?
  does_network_have_names <- if (is_igraph(network))
    !is.null(V(network)$name) else
    !is.null(colnames(network)) || !is.null(rownames(network))
  assert_that(create_names || does_network_have_names)
  # is return_type ok?
  assert_that(is.flag(include_t0))

  # honour create_names: if TRUE, add node names
  if (create_names && !does_network_have_names && is_igraph(network)) {
    V(network)$name <- 1:length(V(network))
  } else if (create_names && !does_network_have_names && !is_igraph(network)) {
    colnames(network) <- 1:ncol(network)
    rownames(network) <- 1:nrow(network)
  }

  # we work with adjacency matrices, so if network is_igraph, convert to an
  # adjacency matrix
  if (is_igraph(network) && is.null(E(network)$weight))
    network <- as_adjacency_matrix(network)
  if (is_igraph(network) && !is.null(E(network)$weight))
    network <- as_adjacency_matrix(network, attr="weight")
  assert_that(all(colnames(network) == rownames(network)))

  # does the adjmat have any negative values (negative edges)?
  assert_that(all(as.matrix(network) >= 0))

  # is there any node in start_run which does not exist in network?
  not_exist_idx <- which(!start_run$node %in% colnames(network))
  not_exist_node <- start_run$node[not_exist_idx]
  not_exist_node <- unique(not_exist_node)
  if (length(not_exist_node) > 0) warning(
    "These nodes specified in start_run don't exist in network: ",
    paste0(not_exist_node, collapse=", "))

  # it is easier if we assume retention is always a vector
  if (length(retention) == 1) retention <- rep(retention, n_nodes)

  # variables in the loop:
  # -   d :: numeric integer vector, the degree of each node
  # - a_t :: numeric vector, the activation at time t
  # - mat :: base::matrix version of network
  d <- Matrix::rowSums(network)
  n_nodes <- length(d)
  a_t <- rep(0, n_nodes) # empty vector
  mat <- as.matrix(network)

  # pre-loop set-up: if !has_name(start_run, "time"), start_run specifies the
  # activation value of certain nodes at t=0. Otherwise, start_run specifies
  # when to add how much activation to certain nodes, so check for activations
  # to be added at t=0.
  if (!has_name(start_run, "time")) start_run$time <- 0

  current_time <- 0
  activations <- numeric(0)
  while (TRUE) {
    # spreading activation
    a_tm1 <- a_t
    mat_t <- create_mat_t(mat, a_tm1, d, retention)
    # create_mat_t creates an "activation matrix", where each (i, j) entry is
    # the activation at time t, at node j, due to node i.
    a_t <- colSums(mat_t)

    # decay and suppress
    a_t <- (1-decay) * a_t
    a_t[a_t < suppress] <- 0

    # apply start_run instructions
    start_run_t <- subset(start_run, time == current_time)
    for (i in 1:nrow(start_run_t)) {
      j <- which(colnames(network) == start_run_t[i, "node"])
      a_t[j] <- a_t[j] + start_run_t[i, "activation"]
    }

    # record
    activations <- c(activations, a_t)

    # check termination
    if (!is.null(time) && current_time >= time) break
    if (!is.null(threshold_to_stop) && all(a_t < threshold_to_stop)) break
    if (is.null(time) && decay == 0 && !never_stop && current_time > 10000) {
      warning(
        "Stopping because there might potentially be an infinite loop. ",
        "Set NEVER_STOP=TRUE to override.")
      break
    }

    current_time <- current_time + 1
  }

  nodes <- rep(colnames(network), current_time+1)
  is <- rep(0:(current_time), rep(n_nodes, current_time+1))
  return_df <- data.frame(
    node=nodes, activation=activations, time=is, stringsAsFactors=TRUE)

  if (!include_t0) {
    return_df <- subset(return_df, time != 0)
    if (nrow(return_df) != 0) rownames(return_df) <- 1:nrow(return_df)
  }

  return_df
}

adj_mat <- matrix(
  c(0, 1,
    0, 0), nrow=2, byrow=TRUE)
initial <- data.frame(node=1, activation=20)

test_that("warning shown if infinite loop", {
  expect_snapshot_error(
    spreadr(adj_mat, initial, retention = 0, time = NULL, threshold_to_stop = 1),
    class="warning")
})

suppressPackageStartupMessages({library(igraph)})

network <- graph_from_adjacency_matrix(matrix(
  c(0, 1, 0, 0, 0, 0, 0, 0,
    0, 0, 1, 0, 0, 0, 0, 0,
    0, 0, 0, 1, 0, 0, 0, 0,
    0, 0, 0, 0, 1, 0, 0, 0,
    0, 0, 0, 0, 0, 1, 0, 0,
    0, 0, 0, 0, 0, 0, 1, 0,
    0, 0, 0, 0, 0, 0, 0, 1,
    1, 0, 0, 0, 0, 0, 0, 0), nrow=8, byrow=TRUE))
V(network)$name <- 1:8
start_run <- data.frame(
  node=as.factor(1),
  activation=100,
  time=c(0, 2, 4, 6))

test_that("error if include_t0 is not flag", {
  expect_snapshot_error(
    spreadr(network, start_run, time=32, retention=0.1, include_t0=42))
})

test_that("include_t0 works well", {
  layout <- layout_in_circle(network)
  results <- spreadr(
    network, start_run, time=16, retention=0.05, include_t0=TRUE)
  results <- suppressMessages(dplyr::left_join(
    results,
    data.frame(node=as.factor(1:8), x=layout[,1], y=layout[,2])))
  expect_snapshot(results)
})

test_that("include_t0 works well in animated GIF", {
  skip_if_not_installed("ggraph")
  skip_if_not_installed("gganimate")
  skip_if_not_installed("gifski")  # gif renderer, requires cargo (as in rust)
  skip_if(
    Sys.getenv("TEST_ANIMATED_GIF") == "",
    paste0(c(
      "The snapshot tests for animated GIFs may fail if you are on a different ",
      "development environment than the one which had generated and committed the ",
      "current GIF snapshots. To continue with this test, set the environment variable",
      "TEST_ANIMATED_GIF to any non-empty string. Then, feel free to overwrite and",
      "commit the current snapshot with the ones you have generated."),
      collapse="\n"))
  suppressPackageStartupMessages({
    library(ggraph)
    library(gganimate)})

  layout <- layout_in_circle(network)
  results <- spreadr(
    network, start_run, time=16, retention=0.05, include_t0=TRUE)
  results <- suppressMessages(dplyr::left_join(
    results,
    data.frame(node=as.factor(1:8), x=layout[,1], y=layout[,2])))

  g <- ggraph(network, layout) +
    # Draw edges with arrows, end_cap is padding between arrow tip and node.
    geom_edge_link(arrow=arrow(), end_cap=circle(5, "mm")) +
    # Draw nodes, including information about activation for transition_manual
    geom_node_point(
      data=results, size=10, mapping=aes(x=x, y=y, colour=activation)) +
    # Label nodes, nudging a bit to the right for aesthetics
    geom_node_label(aes(label=name), nudge_x=.13) +
    # State the parameters of the simulation
    annotate(
      "text", x=-0.65, y=0, hjust=0, label=paste0(
        "data.frame(\n  node=as.factor(1), activation=100, time=c(0,2,4,6))\n",
        "retention=0.05")) +
    # State time point in the title of the plot
    labs(title="t = {current_frame}") +
    transition_manual(time)

  # Silence side-effect prints and messages
  capture.output(
    suppressMessages(
      anim_save("include_t0.gif", g, renderer=gifski_renderer())))

  expect_snapshot_file("include_t0.gif")
})

suppressPackageStartupMessages({library(igraph)})

network <- graph_from_adjacency_matrix(matrix(
  c(0, 1, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 1, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 1, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 1, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0 ,0,
    0, 0, 0, 0, 0, 0, 0, 1, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 1,
    0, 0, 0, 0, 0, 0, 0, 0, 0), nrow=9, byrow=TRUE))
retention <- c(0, 0, 0, 0, 0.5, 0, 0, 1, 0)
V(network)$name <- 1:9
start_run <- data.frame(node=as.factor(c(1, 4, 7)), activation=100, time=0)

test_that("error if length of retention numeric vector is wrong", {
  expect_snapshot_error(
    spreadr(network, start_run, time=5, retention=c(0.5, 0.7), include_t0=TRUE))
})

test_that("per-node retention works well", {
  layout <- layout_on_grid(network)
  results <- spreadr(
    network, start_run, time=5, retention=retention, include_t0=TRUE)
  results <- suppressMessages(dplyr::left_join(
    results, data.frame(node=as.factor(1:9), x=layout[,1], y=layout[,2])))
  expect_snapshot(results)
})

test_that("per-node retention works well with animated GIF", {
  skip_if_not_installed("ggraph")
  skip_if_not_installed("gganimate")
  skip_if_not_installed("gifski")  # gif renderer, requires cargo (as in rust)
  skip_if(
    Sys.getenv("TEST_ANIMATED_GIF") == "",
    paste0(c(
      "The snapshot tests for animated GIFs may fail if you are on a different ",
      "development environment than the one which had generated and committed the ",
      "current GIF snapshots. To continue with this test, set the environment variable",
      "TEST_ANIMATED_GIF to any non-empty string. Then, feel free to overwrite and",
      "commit the current snapshot with the ones you have generated."),
      collapse="\n"))
  suppressPackageStartupMessages({
    library(ggraph)
    library(gganimate)})

  layout <- layout_on_grid(network)
  results <- spreadr(
    network, start_run, time=5, retention=retention, include_t0=TRUE)
  results <- suppressMessages(dplyr::left_join(
    results, data.frame(node=as.factor(1:9), x=layout[,1], y=layout[,2])))

  g <- ggraph(network, layout) +
    # Draw edges with arrows, end_cap is padding between arrow tip and node.
    geom_edge_link(arrow=arrow(), end_cap=circle(5, "mm")) +
    # Draw nodes, including information about activation for transition_manual
    geom_node_point(
      data=results, size=10, mapping=aes(x=x, y=y, colour=activation)) +
    # Label nodes, nudging a bit to the right for aesthetics
    geom_node_label(aes(label=name), nudge_x=.13) +
    # State the parameters of the simulation. Nodes 1, 2, 3 are at the bottom.
    # 7, 8, 9 at the top.
    annotate("text", x=0.1, y=2.1, hjust=0, label="retentions: c(0, 1, 0)") +
    annotate("text", x=0.1, y=1.1, hjust=0, label="retentions: c(0, 0.5, 0)") +
    annotate("text", x=0.1, y=0.1, hjust=0, label="retentions: c(0, 0, 0)") +
    # State time point in the title of the plot
    labs(title="t = {current_frame}") +
    transition_manual(time)

  # Silence side-effect prints and messages
  capture.output(
    suppressMessages(
      anim_save("per-node-retention.gif", g, renderer=gifski_renderer())))

  expect_snapshot_file("per-node-retention.gif")
})

# Here, we test simulations from:
# Siew, Cynthia S. Q. 2019. ‘Spreadr: An R Package to Simulate Spreading
# Activation in a Network’. Behavior Research Methods 51 (2): 910–29.
# https://doi.org/10.3758/s13428-018-1186-5.

library(fs)
skip_on_cran()  # Can't seem to download stuff on CRAN

# We download the test datasets and expected output from spreadr's OSF
# repository at https://osf.io/a9bv6/. The downloaded files are excluded from
# the package tarball with .Rbuildignore.
download_urls <- list(
  # Lexical retrieval simulation
  "ego2hopnets_24.RData"    = "https://osf.io/download/vucnf/",
  "study1_rawoutput.zip"    = "https://osf.io/download/vy42b/",
  "24toynets.csv"           = "https://osf.io/download/9wz6s/",
  # False memory simulation
  "study2_rawoutput.zip"    = "https://osf.io/download/6xzgq/",
  # Semantic priming simulation
  "usf.RData"               = "https://osf.io/download/gwyc8/",
  "spp_200_pairs.RData"     = "https://osf.io/download/h42ay/",
  # Because these simulations run for many iterations (800), each on a rather
  # large network (10k nodes, 63k edges), they will take a /very/ long time to
  # test. Instead, we prepare to test only the first pair of words, for one
  # retention value of 0.8, and we really run that test only if the user says
  # to do so, by specifying an environment variable TEST_SEMANTIC_PRIMING to
  # any non-empty string.
  # "spp_retention_0.2.RData" = "https://osf.io/download/yjt4g/",
  # "spp_retention_0.4.RData" = "https://osf.io/download/f48th/",
  # "spp_retention_0.6.RData" = "https://osf.io/download/uk6rz/",
  "spp_retention_0.8.RData" = "https://osf.io/download/ybq6u/")
# Download files
sapply(names(download_urls), function(fn) {
  if (!file_exists(fn)) download.file(download_urls[[fn]], fn, quiet=TRUE)
})
# Extract files
zip_files <- Filter(function(fn) grepl("\\.zip$", fn), names(download_urls))
sapply(zip_files, function(zfn) {
  if (!dir_exists(gsub("\\.zip$", "", zfn))) unzip(zfn)
})

load("ego2hopnets_24.RData")
v2011 <- read.csv('24toynets.csv')
retentions <- seq(0.1, 0.9, by=0.1)

for (retention in retentions) {
  for (i in 1:nrow(v2011)) {
    test_that(sprintf(
      "Lexical retrieval simulation: network %d, retention %f", i, retention), {
        results <- spreadr(
          start_run=data.frame(node=v2011$name[i], activation=100),
          network=ego2hopnets_24[[i]],
          decay=0, retention=retention, suppress=0, time=10)
        expected <- read.csv(
          sprintf(
            "study1_rawoutput/%d_%s_%.1f.csv", i, v2011$name[i], retention),
          stringsAsFactors=TRUE,
          colClasses=c("NULL", NA, NA, NA))  # ignore the first column
        expect_equal(results, expected)
      }
    )
    test_that(sprintf(
      "False memory simulation: network %d, retention %f", i, retention), {
        neighbour_nodes <- igraph::neighbors(
          ego2hopnets_24[[i]], as.character(v2011$name[i]))
        results <- spreadr(
          start_run=data.frame(
            node=as.vector(neighbour_nodes$name),
            activation=100/length(neighbour_nodes),
            stringsAsFactors=FALSE),
          network=ego2hopnets_24[[i]],
          decay=0, retention=retention, suppress=0, time=10)
        expected <- read.csv(
          sprintf(
            "study2_rawoutput/%d_%s_%.1f_sim2.csv", i, v2011$name[i], retention),
          stringsAsFactors=TRUE,
          colClasses=c("NULL", NA, NA, NA))  # ignore the first column
        expect_equal(results, expected)
      }
    )
  }
}

load("usf.RData")
usf <- igraph::simplify(usf)
igraph::E(usf)$weight <- 1
load("spp_200_pairs.RData")
load("spp_retention_0.8.RData")  # loaded into name x
spp_retention_0.8 <- x

test_that("Semantic priming simulation: pair 1, retention 0.8", {
  skip_if(
    Sys.getenv("TEST_SEMANTIC_PRIMING") == "",
    paste0(c(
      "This test takes a few minutes, during which it will use up all system memory",
      "on most personal computers and laptops. If you are sure you would like to run",
      "this test, set the environment variable TEST_SEMANTIC_PRIMING to any non-empty ",
      "string."),
      collapse="\n"))
  start_run <- data.frame(
    node=spp_200_pairs$Prime[1], activation=100, stringsAsFactors=FALSE)
  result <- spreadr(
    start_run=start_run, network=usf,
    retention=0.8, time=10, suppress=0, decay=0)
  targets <- subset(result, node == spp_200_pairs$TargetWord[1])
  result_vector <- c(
    spp_200_pairs$Prime[1], spp_200_pairs$TargetWord[1], targets$activation)
  expected_vector <- spp_retention_0.8[1, ]
  expect_equal(result_vector, expected_vector)
  gc()
})

adj_mat <- matrix(
  c(0, 1,
    0, 0), nrow=2, byrow=TRUE)

test_that("start_run without time column assumes t=0", {
  results <- spreadr(
    adj_mat, data.frame(node=as.factor(1:2), activation=c(10, 10)), time=3)
  expected <- data.frame(
    node=as.factor(c(1,  2,   1,    2,    1,     2)),
    activation=c(    5, 15, 2.5, 17.5, 1.25, 18.75),
    time      =c(    1,  1,   2,    2,    3,     3))
  expect_equal(results, expected)
})

test_that("start_run with time column adds activation at the right times", {
  results <- spreadr(
    adj_mat,
    data.frame(node=as.factor(c(1, 1)), activation=c(10, 10), time=1:2),
    time=3)
  expected <- data.frame(
    node=as.factor(c( 1, 2,  1, 2,   1,    2)),
    activation=c(    10, 0, 15, 5, 7.5, 12.5),
    time      =c(     1, 1,  2, 2,   3,    3))
  expect_equal(results, expected)
})

# See https://github.com/csqsiew/spreadr/issues/7.

library(igraph)

test_that("error when edges are negative", {
  g <- graph_from_literal(1--2, 1--3, 1--4)
  E(g)$weight <- c(-1, -2, -1) # negative weights
  expect_snapshot_error(
    spreadr(g, data.frame(node = 1, activation = 100), time = 1)
  )
})

adj_mat <- matrix(
  c(0, 1,
    0, 0), nrow=2, byrow=TRUE)
initial <- data.frame(node=1, activation=20)

test_that("error thrown if both time, threshold_to_stop is NULL", {
  expect_error(
    spreadr(adj_mat, initial, time=NULL),  # threshold_to_stop defaults to NULL
    "time and threshold_to_stop cannot both be NULL")
})

test_that("terminates with time only", {
  results <- spreadr(adj_mat, initial, time=1)
  expected <- data.frame(
    node=as.factor(1:2),
    activation=c(10, 10),
    time=c(1, 1))
  expect_equal(results, expected)
})

test_that("terminates with threshold_to_stop only", {
  results <- spreadr(adj_mat, initial, time=NULL, threshold_to_stop=2.5, decay=0.5)
  expected <- data.frame(
    node=as.factor(rep(1:2, 3)),
    activation=c(
      20/4,                20/2/2,                    # t=1
      20/4/4,    (20/2/2 + 20/4/2)/2,                 # t=2
      20/4/4/4, ((20/2/2 + 20/4/2)/2 + 20/4/4/2)/2),  # t=3
    time=c(1, 1, 2, 2, 3, 3))
  expect_equal(results, expected)
})

test_that("terminates with both time and threshold_to_stop at threshold", {
  results <- spreadr(adj_mat, initial, time=6, threshold_to_stop=2.5, decay=0.5)
  expected <- data.frame(
    node=as.factor(rep(1:2, 3)),
    activation=c(
      20/4,                20/2/2,                    # t=1
      20/4/4,    (20/2/2 + 20/4/2)/2,                 # t=2
      20/4/4/4, ((20/2/2 + 20/4/2)/2 + 20/4/4/2)/2),  # t=3
    time=c(1, 1, 2, 2, 3, 3))
  expect_equal(results, expected)
})

test_that("terminates with both time and threshold_to_stop at time", {
  results <- spreadr(adj_mat, initial, time=4, threshold_to_stop=1)
  expected <- data.frame(
    node=as.factor(rep(1:2, 4)),
    activation=c(
    # t=  1,    2,        3,          4
      10,10, 5,15, 2.5,17.5, 1.25,18.75),
    time=c(1, 1, 2, 2, 3, 3, 4, 4))
  expect_equal(results, expected)
})

# See https://github.com/csqsiew/spreadr/issues/4.

test_that("Weighted graph, when represented by a matrix, works", {
  weighted_network <- matrix(
    c(0, 1, 9,
      1, 0, 0,
      9, 0, 0), nrow=3, byrow=TRUE)
  colnames(weighted_network) <- c("a", "b", "c")
  rownames(weighted_network) <- c("a", "b", "c")
  results <- spreadr(
    weighted_network, data.frame(node="a", activation=10),
    time=1, retention=0, include_t0=TRUE)
  expected <- data.frame(
    node=as.factor(rep(c("a", "b", "c"), 2)),
    activation=c(10, 0, 0, 0, 1, 9),
    time=c(rep(0, 3), rep(1, 3)))
  expect_equal(results, expected)
})

test_that("GitHub issue 1 is fixed", {
  # See: https://github.com/csqsiew/spreadr/issues/1. Note that the input
  # adjacency matrix is different here. When the issue was written, the input
  # adjacency matrix had the (i, j) entry representing the edge from j to i
  # Now, the (i, j) entry represents the edge from i to j instead.
  adj_mat <- matrix(nrow=5, ncol=5, data=c(
    0, 0, 0, 1, 0,
    1, 0, 0, 1, 0,
    0, 0, 0, 1, 0,
    0, 1, 0, 0, 0,
    1, 1, 0, 0, 0))
  results <- spreadr(
    adj_mat, data.frame(node=1, activation=20),
    retention=0, time=3, suppress=0, decay=0)
  expected <- data.frame(
    node=as.factor(rep(1:5, 3)),
    activation=c(
      0, 10, 0, 0, 10,
      0, 0, 0, 5, 15,
      5/3, 5/3, 5/3, 0, 15),
    time=c(rep(1, 5), rep(2, 5), rep(3, 5)))

  expect_equal(results, expected)
})

---
title: "spreadr: A R package to simulate spreading activation in a network"
author: "Cynthia S. Q. Siew"
date: "`r Sys.Date()`"
output:
  html_document:
    toc: true
    toc_float: true
bibliography: sources.bib
vignette: >
  %\VignetteIndexEntry{Introduction}
  %\VignetteEngine{knitr::rmarkdown}
  \usepackage[utf8]{inputenc}
---

<style>
p { text-align: justify }
</style>

# Introduction

The notion of spreading activation is a prevalent metaphor in the cognitive sciences; however, the tools to implement spreading activation in a computational simulation are not as readily available.
This vignette introduces the spreadr R package (pronunced 'SPREAD-er'), which can implement spreading activation within a specified network structure.
The algorithmic method implemented in the `spreadr` function follows the approach described in @vitevitchSimulatingRetrievalHighly2011, who viewed activation as a fixed cognitive resource that could “spread” among connected nodes in a network.

# Installation {.tabset}

You can choose to install the stable version from CRAN, or the development build from GitHub.

## Stable (CRAN)

```{r eval=FALSE}
install.packages("spreadr")
```

If you encounter any bugs or issues, please try the development build first.
The bug or issue may have already been fixed on GitHub, but not yet propagated onto CRAN.

## Development (GitHub)

```{r eval=FALSE}
install.packages("remotes")
remotes::install_github("csqsiew/spreadr")
```

If you encounter any bugs or issues, please try the development build first.
The bug or issue may have already been fixed on GitHub, but not yet propagated onto CRAN.

# Example with a phonological network

In this example, we will simulate spreading activation in a small sample portion of a phonological network [@chanInfluencePhonologicalNeighborhood2009] which is automatically loaded with spreadr.
This phonological network is unweighted and undirected, but spreadr supports weighted and directed graphs as well.
This makes it is possible to simulate spreading activation in a weighted network where more activation is passed between nodes that have "stronger" edges, or in a directed (asymmetric) network where activation can pass from node *i* to node *j* but not necessarily from node *j* to node *i*.

You may substitute any network in place of this example one.

## Describe the network {.tabset}

The network for spreading activation must be either an `igraph` object or an adjacency matrix.

### Using an igraph object

`pnet`, an `igraph` object with named vertices representing our sample phonological network is automatically loaded with the spreadr library.

```{r message=FALSE}
library(spreadr)
library(igraph)

data("pnet")  # load and inspect the igraph object
```

```{r eval=FALSE}
plot(pnet)
```
<!-- Behind the scenes, we want to generate identical plots, so save a layout.
     This is hidden because I don't want to make the vignette too complex. -->
```{r echo=FALSE}
set.seed(1)
layout <- layout_with_lgl(pnet)
plot(pnet, layout=layout)
```

Don't worry if your plot does not look exactly as above.
The layout of the nodes is determined stochastically.

### Using an adjacency matrix

`pnetm`, an named adjacency matrix representing our sample phonological network is automatically loaded with the spreadr library.

```{r message=FALSE}
library(spreadr)
library(igraph)
set.seed(1)

data("pnetm")
pnetm[1:5, 1:5]  # inspect the first few entries
```

```{r eval=FALSE}
plot(graph_from_adjacency_matrix(
  pnetm, mode="undirected")) # visualise the graph
```
```{r echo=FALSE}
plot(graph_from_adjacency_matrix(
    pnetm, mode="undirected"),
  layout=layout)
```

For those following along: don't worry if your plot does not look exactly as above.
The layout of the nodes is determined stochastically.

For simplicity, the rest of this example will use `pnet`, an `igraph` representation of `pnetm`.
This difference is trivial --- the spreadr function accepts both igraph and adjacency matrix.

## Specify the adding of activation {.tabset}

A simulation of spreading activation is uninteresting without some activation to spread.
In the simplest case, you may specify the initial activation state of each node, from which the simulation will proceed.
Alternatively, you can also specify the addition of activation at any node, at any time point within the simulation.

### Initial activation only

Initial activation is specified by a `data.frame` (or `data.frame`-like object, such as a `tibble`) with columns `node` and `activation`.
Each row represents the addition of `activation` amount of activation to the node with name `node`, at the initial pre-spreading activation step of the simulation.

For example, if we wanted the nodes `"beach"` and `"speck"` to have `20` and `10` activation initially, respectively:

```{r}
start_run <- data.frame(
  node=      c("beach", "speck"),
  activation=c(     20,      10))
```

### Activation at specified time points

The adding of activation at specified time points is specified by a `data.frame` (or `data.frame`-like object, such as a `tibble`) with columns `node`, `activation`, and `time`.
Each row represents the addition of `activation` amount of activation to the node with name `node` at time point `time`.
The initial state before any spreading activation occurs is `time = 0`.

Therefore, if we wanted the node `"beach"` to have `20` activation initially, then `"speck"` to have `10` activation at `time = 5`:

```{r}
start_run <- data.frame(
  node =       c("beach", "speck"),
  activation = c(     20,      10),
  time =       c(      0,       5))
```

For more details about the meaning of `time` in the spreading activation simulation, please read the `spreadr` function documentation.
To keep things simple, the rest of this example will use the version of `start_run` without the `time` column (see the leftmost tab).

```{r echo=FALSE}
start_run <- data.frame(
  node =       c("beach", "speck"),
  activation = c(     20,      10))
```

## Run the simulation

We are now ready to run the simulation through the `spreadr` function.
This function takes a number of parameters, some of which are listed here.
Required parameters are written in **bold**, optional parameters are written with their default values (in brackets):

i. **network**: Adjacency matrix or `igraph` object representing the network in which to simulate spreading activation
ii. **start_run**: Non-empty `data.frame` describing the addition of activation into the network.
iii. *decay* (`0`): Proportion of activation lost at each time step (range from 0 to 1)
iv. *retention* (`0.5`): Proportion of activation retained in the originator node (range from 0 to 1)
v. *suppress* (`0`): Nodes with activation values lower than this value will have their activations forced to `0` (typically this will be a very small value e.g. < 0.001)
vi. *time* (`10`): Number of iterations in the simulation
vii. *include_t0* (`FALSE`): If the initial state activations should be included in the result `data.frame`

For more details on all the possible function parameters, please read the `spreadr` function documentation.
For more details on the spreading activation algorithm, see @siewSpreadrPackageSimulate2019 and @vitevitchSimulatingRetrievalHighly2011.

```{r}
result <- spreadr(pnet, start_run, include_t0=TRUE)
```

## Examine the results

The result is presented as a `data.frame` with columns `node`, `activation`, and `time`.
Each row contains the activation value of a node at its specified time step.

```{r eval=TRUE}
head(result)  # inspect the first few rows
tail(result)  # inspect the last few rows
```

As a `data.frame`, the `result` can be easily saved as a CSV file for sharing.

```{r eval=FALSE}
write.csv(result, file="result.csv")  # save result as CSV file
```

You can also easily visualise the `result` using your favourite graphical libraries.
Here, we use `ggplot2`:

```{r}
library(ggplot2)
ggplot(result, aes(x=time, y=activation, color=node)) +
  geom_point() +
  geom_line()
```

# Additional examples

Here, we address some common "how do I do *X*?" questions by example.

## Using a weighted network

Simply pass a weighted network (`igraph` object or adjacency matrix) to the `spreadr` function.

Let's take a simple three-node network as example:

```{r}
weighted_network <- matrix(
  c(0, 1, 9,
    1, 0, 0,
    9, 0, 0), nrow=3, byrow=TRUE)
colnames(weighted_network) <- c("a", "b", "c")
rownames(weighted_network) <- c("a", "b", "c")

# To visualise the network only --- this is not necessary for spreadr
weighted_igraph <- graph_from_adjacency_matrix(
  weighted_network, mode="undirected", weighted=TRUE)
plot(weighted_igraph, edge.width=E(weighted_igraph)$weight)
```

If we let the node `"a"` have initial `activation = 10`, we should expect to see 9 units of activation go to `"c"`, and 1 unit to `"b"`, for `retention = 0`.

```{r}
spreadr(
  weighted_igraph, data.frame(node="a", activation=10),
  time=1, retention=0, include_t0=TRUE)
```

## Using a directed network

Simply pass a directed network (`igraph` object or adjacency matrix) to the `spreadr` function.

Let's take a simple three-node network as example:

```{r}
directed_network <- matrix(
  c(0, 1, 0,
    0, 0, 1,
    0, 0, 0), nrow=3, byrow=TRUE)
colnames(directed_network) <- c("a", "b", "c")
rownames(directed_network) <- c("a", "b", "c")

# To visualise the network only --- this is not necessary for spreadr
directed_igraph <- graph_from_adjacency_matrix(
  directed_network, mode="directed")
plot(directed_igraph, edge.width=E(directed_igraph)$weight)
```

If we let the node `"b"` have initial `activation = 10`, we should expect to see all 10 units of activation go to `"c"`, and none go to `"a"`, for `retention = 0`.

```{r}
spreadr(
  directed_igraph, data.frame(node="b", activation=10),
  time=1, retention=0, include_t0=TRUE)
```

## Repeating the simulation across different parameters

There are a few ways to do this, but here we will showcase a relatively simple method using `data.frame`s.

Suppose we want to simulate spreading activation across four different parameter sets of *(retention, decay*): (0, 0), (0.5, 0), (0, 0.5), and (0.5, 0.5).
We would first record down all those parameters which will differ into a `data.frame`:

```{r}
params <- data.frame(
  retention=c(0, 0.5,   0, 0.5),
  decay=    c(0,   0, 0.5, 0.5))
```

Then, we prepare the common parameters that will *not* differ.

```{r}
network <- matrix(
  c(0, 1,
    0, 0), nrow=2, byrow=TRUE)
start_run <- data.frame(node=1, activation=10)
```

Now, to run the simulate once for each set of different parameters, we simply `apply` over the rows of `params`.

```{r}
apply(params, 1, function(row)
  spreadr(
    network, start_run,
    time=2, include_t0=TRUE,
    retention=row[1], decay=row[2]))
```

# Reporting bugs or issues

First, please check that you are on the development build.
This is because the development build contains a more up-to-date version of spreadr, which may include various bug fixes, as compared to the stable CRAN build.

If your problems persist, please report them on our [Github repository](https://github.com/csqsiew/spreadr/issues).

# Session Info

```{r}
sessionInfo()
```

# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spreadr.R
\name{spreadr}
\alias{spreadr}
\title{Simulate spreading activation in a network}
\usage{
spreadr(
  network,
  start_run,
  retention = 0.5,
  time = 10,
  threshold_to_stop = NULL,
  decay = 0,
  suppress = 0,
  include_t0 = FALSE,
  create_names = TRUE,
  never_stop = FALSE
)
}
\arguments{
\item{network}{Adjacency matrix or
\code{\link[igraph:igraph-package]{igraph}} object representing the network
in which to simulate spreading activation.}

\item{start_run}{Non-empty \code{\link[base]{data.frame}} with mandatory
columns \emph{node}, \emph{activation}; and optional columns \emph{time}.
If the \emph{time} column is present, \emph{activation} is added to
\emph{node} at each \emph{time}. Otherwise, the \emph{activation}s are
added to their corresponding nodes at \code{t = 0}.}

\item{retention}{Number from 0 to 1 (inclusive) or a numeric vector of such
numbers of length equals number of nodes in the network. This represents
the proportion of activation that remains in the node (not spread) at each
time step. Then, \code{1 - retention} of the activation at each node is
spread to neighbouring nodes. If a numeric vector, retentions are assigned
to nodes according to the order given by \code{V(network)} if
\code{network} is an \code{\link[igraph:igraph-package]{igraph}} object or
\code{nrow(network)} if \code{network} is an adjacency matrix.}

\item{time}{Positive non-zero integer, or \code{NULL}. If not \code{NULL},
the number of time steps to simulate before stopping. Otherwise,
stop with the \code{threshold_to_stop} parameter.}

\item{threshold_to_stop}{Number or \code{NULL}. If not \code{NULL}, stop the
simulation only when all nodes have activation value less than
\code{threshold_to_stop}. Otherwise, stop with the \code{time} parameter.}

\item{decay}{Number from 0 to 1 (inclusive) representing the proportion of
activation that is lost at each time step.}

\item{suppress}{Number representing the maximum amount of activation in a
node for it to be set to 0, at each time step.}

\item{include_t0}{Boolean flag indicating if activation at \code{t = 0}
should be prepended to the output \code{data.frame}. This is \code{FALSE}
by default for back-compatibility.}

\item{create_names}{Boolean flag indicating if nodes should be automatically
named (\code{1:n}, where \code{n} is the number of nodes) in case they are
missing.}

\item{never_stop}{Boolean flag indicating if the simulation should be stopped
if there have been too many iterations (so that there might be an infinite
loop).}
}
\value{
A \code{\link[base]{data.frame}} with \emph{node}, \emph{activation}
  and \emph{time} columns representing the spread of activation in the network
  over time.
}
\description{
Simulate spreading activation in a network
}
\details{
At least one of parameters \code{time} or \code{threshold_to_stop}
  must be non-\code{NULL}. If both are non-\code{NULL}, the simulation stops
  at the earliest time possible.

  The simulation iterates like so: for every \code{i} in \code{[0, time]},
  \itemize{
    \item{Spread activation from node to node}
    \item{Decay the activation at each node by the proportion specified by
      \code{decay}}
    \item{Set the activation at nodes with activation less than
      \code{suppress} to \code{0}}
    \item{Add the activations in \code{start_run} with \code{time = i} to
      their corresponding nodes}
    \item{Save the activations at each node for output}
    \item{Check the terminating conditions \code{time} and
      \code{threshold_to_stop}. If any are satisfied, terminate the
      simulation.}
  }
}
\examples{
# make an adjacency matrix and randomly fill some cells with 1s
mat <- matrix(sample(c(0,1), 100, replace=TRUE), 10, 10)
diag(mat) <- 0 # remove self-loops
initial_df <- data.frame(node=1, activation=20, stringsAsFactors=FALSE)
results <- spreadr(mat, initial_df)

head(results, 10)
tail(results, 10)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{pnetm}
\alias{pnetm}
\title{Small example of a phonological network as an adjacency matrix}
\format{
Adjacency matrix representing an unweighted undirected graph with
  34 vertices and 96 edges. There are no self-loops.
}
\source{
Ying, Chan & Vitevitch, Michael. (2009). The Influence of the
  Phonological Neighborhood Clustering Coefficient on Spoken Word
  Recognition. Journal of experimental psychology. Human perception and
  performance. 35. 1934-49. 10.1037/a0016902.
}
\usage{
pnetm
}
\description{
Small example of a phonological network as an adjacency matrix
}
\keyword{datasets}
